/*
 * Shadowsocks - A shadowsocks client for Android
 * Copyright (C) 2014 <max.c.lv@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *                            ___====-_  _-====___
 *                      _--^^^#####//      \\#####^^^--_
 *                   _-^##########// (    ) \\##########^-_
 *                  -############//  |\^^/|  \\############-
 *                _/############//   (@::@)   \\############\_
 *               /#############((     \\//     ))#############\
 *              -###############\\    (oo)    //###############-
 *             -#################\\  / VV \  //#################-
 *            -###################\\/      \//###################-
 *           _#/|##########/\######(   /\   )######/\##########|\#_
 *           |/ |#/\#/\#/\/  \#/\##\  |  |  /##/\#/  \/\#/\#/\#| \|
 *           `  |/  V  V  `   V  \#\| |  | |/#/  V   '  V  V  \|  '
 *              `   `  `      `   / | |  | | \   '      '  '   '
 *                               (  | |  | |  )
 *                              __\ | |  | | /__
 *                             (vvv(VVV)(VVV)vvv)
 *
 *                              HERE BE DRAGONS
 *
 */

package com.github.shadowsocks

import java.io.File
import java.net.{Inet6Address, InetAddress}
import java.util.Locale

import android.app._
import android.content._
import android.content.pm.{PackageInfo, PackageManager}
import android.net.{ConnectivityManager, Network}
import android.os._
import android.support.v4.app.NotificationCompat
import android.support.v4.content.ContextCompat
import android.util.{Log, SparseArray}
import android.widget.Toast
import com.github.shadowsocks.aidl.Config
import com.github.shadowsocks.utils._

import scala.collection._
import scala.collection.mutable.ArrayBuffer
import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.Future

class ShadowsocksNatService extends Service with BaseService {

  val TAG = "ShadowsocksNatService"

  val CMD_IPTABLES_RETURN = " -t nat -A OUTPUT -p tcp -d 0.0.0.0 -j RETURN"
  val CMD_IPTABLES_DNAT_ADD_SOCKS = " -t nat -A OUTPUT -p tcp " +
    "-j DNAT --to-destination 127.0.0.1:8123"

  var lockReceiver: BroadcastReceiver = null
  var closeReceiver: BroadcastReceiver = null
  var connReceiver: BroadcastReceiver = null
  var config: Config = null
  var apps: Array[ProxiedApp] = null
  val myUid = Process.myUid()

  var sslocalPid: Int = 0;
  var sstunnelPid: Int = 0;
  var redsocksPid: Int = 0;
  var pdnsdPid: Int = 0;

  private val dnsAddressCache = new SparseArray[String]

  def getNetId(network: Network): Int = {
    network.getClass.getDeclaredField("netId").get(network).asInstanceOf[Int]
  }

  def restoreDnsForAllNetwork() {
    val manager = getSystemService(Context.CONNECTIVITY_SERVICE).asInstanceOf[ConnectivityManager]
    val networks = manager.getAllNetworks
    val cmdBuf = new ArrayBuffer[String]()
    networks.foreach(network => {
      val netId = getNetId(network)
      val oldDns = dnsAddressCache.get(netId)
      if (oldDns != null) {
        cmdBuf.append("ndc resolver setnetdns %d \"\" %s".formatLocal(Locale.ENGLISH, netId, oldDns))
        dnsAddressCache.remove(netId)
      }
    })
    if (cmdBuf.nonEmpty) Console.runRootCommand(cmdBuf.toArray)
  }

  def setDnsForAllNetwork(dns: String) {
    val manager = getSystemService(Context.CONNECTIVITY_SERVICE).asInstanceOf[ConnectivityManager]
    val networks = manager.getAllNetworks
    if (networks == null) return

    val cmdBuf = new ArrayBuffer[String]()
    networks.foreach(network => {
      val networkInfo = manager.getNetworkInfo(network)
      if (networkInfo == null) return
      if (networkInfo.isConnected) {
        val netId = getNetId(network)
        val curDnsList = manager.getLinkProperties(network).getDnsServers
        if (curDnsList != null) {
          import scala.collection.JavaConverters._
          val curDns = curDnsList.asScala.map(ip => ip.getHostAddress).mkString(" ")
          if (curDns != dns) {
            dnsAddressCache.put(netId, curDns)
            cmdBuf.append("ndc resolver setnetdns %d \"\" %s".formatLocal(Locale.ENGLISH, netId, dns))
          }
        }
      }
    })
    if (cmdBuf.nonEmpty) Console.runRootCommand(cmdBuf.toArray)
  }

  def setupDns() {
    setDnsForAllNetwork("127.0.0.1")
  }

  def resetDns() = {
    restoreDnsForAllNetwork()
  }

  def flushDns() {
    if (Utils.isLollipopOrAbove) {
      val manager = getSystemService(Context.CONNECTIVITY_SERVICE).asInstanceOf[ConnectivityManager]
      val networks = manager.getAllNetworks
      val cmdBuf = new ArrayBuffer[String]()
      networks.foreach(network => {
        val networkInfo = manager.getNetworkInfo(network)
        if (networkInfo.isAvailable) {
          val netId = network.getClass.getDeclaredField("netId").get(network).asInstanceOf[Int]
          cmdBuf.append("ndc resolver flushnet %d".formatLocal(Locale.ENGLISH, netId))
        }
      })
      Console.runRootCommand(cmdBuf.toArray)
    } else {
      Console.runRootCommand(Array("ndc resolver flushdefaultif", "ndc resolver flushif wlan0"))
    }
  }


  def destroyConnectionReceiver() {
    if (connReceiver != null) {
      unregisterReceiver(connReceiver)
      connReceiver = null
    }
    resetDns()
  }

  def initConnectionReceiver() {
    val filter = new IntentFilter(ConnectivityManager.CONNECTIVITY_ACTION)
    connReceiver = (context: Context, intent: Intent) => setupDns()
    registerReceiver(connReceiver, filter)
  }

  def startShadowsocksDaemon() {
    if (config.route != Route.ALL) {
      val acl: Array[String] = config.route match {
        case Route.BYPASS_LAN => getResources.getStringArray(R.array.private_route)
        case Route.BYPASS_CHN => getResources.getStringArray(R.array.chn_route_full)
      }
      ConfigUtils.printToFile(new File(Path.BASE + "acl.list"))(p => {
        acl.foreach(item => p.println(item))
      })
    }

    val conf = ConfigUtils
      .SHADOWSOCKS.formatLocal(Locale.ENGLISH, config.proxy, config.remotePort, config.localPort,
        config.sitekey, config.encMethod, 600)
    ConfigUtils.printToFile(new File(Path.BASE + "ss-local-nat.conf"))(p => {
      p.println(conf)
    })

    val cmd = new ArrayBuffer[String]
    cmd += (Path.BASE + "ss-local"
          , "-b" , "127.0.0.1"
          , "-t" , "600"
          , "-c" , Path.BASE + "ss-local-nat.conf")

    if (config.isAuth) cmd += "-A"

    if (config.route != Route.ALL) {
      cmd += "--acl"
      cmd += (Path.BASE + "acl.list")
    }

    if (BuildConfig.DEBUG) Log.d(TAG, cmd.mkString(" "))
    sslocalPid = System.exec(cmd.mkString(" "))
  }

  def startTunnel() {
    if (config.isUdpDns) {
      val conf = ConfigUtils
        .SHADOWSOCKS.formatLocal(Locale.ENGLISH, config.proxy, config.remotePort, 8153,
          config.sitekey, config.encMethod, 10)
      ConfigUtils.printToFile(new File(Path.BASE + "ss-tunnel-nat.conf"))(p => {
        p.println(conf)
      })
      val cmd = new ArrayBuffer[String]
      cmd += (Path.BASE + "ss-tunnel"
        , "-u"
        , "-t" , "10"
        , "-b" , "127.0.0.1"
        , "-L" , "8.8.8.8:53"
        , "-c" , Path.BASE + "ss-tunnel-nat.conf")

      cmd += ("-l" , "8153")

      if (config.isAuth) cmd += "-A"

      if (BuildConfig.DEBUG) Log.d(TAG, cmd.mkString(" "))

     sstunnelPid =  System.exec(cmd.mkString(" "))

    } else {
      val conf = ConfigUtils
        .SHADOWSOCKS.formatLocal(Locale.ENGLISH, config.proxy, config.remotePort, 8163,
          config.sitekey, config.encMethod, 10)
      ConfigUtils.printToFile(new File(Path.BASE + "ss-tunnel-nat.conf"))(p => {
        p.println(conf)
      })
      val cmdBuf = new ArrayBuffer[String]
      cmdBuf += (Path.BASE + "ss-tunnel"
        , "-u"
        , "-t" , "10"
        , "-b" , "127.0.0.1"
        , "-l" , "8163"
        , "-L" , "8.8.8.8:53"
        , "-c" , Path.BASE + "ss-tunnel-nat.conf")

      if (config.isAuth) cmdBuf += "-A"

      if (BuildConfig.DEBUG) Log.d(TAG, cmdBuf.mkString(" "))

      sstunnelPid = System.exec(cmdBuf.mkString(" "))
    }
  }

  def startDnsDaemon() {

    val conf = if (config.route == Route.BYPASS_CHN) {
      val reject = ConfigUtils.getRejectList(getContext)
      val blackList = ConfigUtils.getBlackList(getContext)
      ConfigUtils.PDNSD_DIRECT.formatLocal(Locale.ENGLISH, "127.0.0.1", 8153,
        reject, blackList, 8163, "")
    } else {
      ConfigUtils.PDNSD_LOCAL.formatLocal(Locale.ENGLISH, "127.0.0.1", 8153,
        8163, "")
    }

    ConfigUtils.printToFile(new File(Path.BASE + "pdnsd-nat.conf"))(p => {
       p.println(conf)
    })
    val cmd = Path.BASE + "pdnsd -c " + Path.BASE + "pdnsd-nat.conf"

    if (BuildConfig.DEBUG) Log.d(TAG, cmd)

    pdnsdPid = System.exec(cmd)

  }

  def getVersionName: String = {
    var version: String = null
    try {
      val pi: PackageInfo = getPackageManager.getPackageInfo(getPackageName, 0)
      version = pi.versionName
    } catch {
      case e: PackageManager.NameNotFoundException =>
        version = "Package name not found"
    }
    version
  }

  def startRedsocksDaemon() {
    val conf = ConfigUtils.REDSOCKS.formatLocal(Locale.ENGLISH, config.localPort)
    val cmd = Path.BASE + "redsocks -c %sredsocks-nat.conf"
      .formatLocal(Locale.ENGLISH, Path.BASE, Path.BASE)
    ConfigUtils.printToFile(new File(Path.BASE + "redsocks-nat.conf"))(p => {
      p.println(conf)
    })

    if (BuildConfig.DEBUG) Log.d(TAG, cmd)
    redsocksPid = System.exec(cmd)
  }

  /** Called when the activity is first created. */
  def handleConnection: Boolean = {

    startTunnel()
    if (!config.isUdpDns) startDnsDaemon()
    startRedsocksDaemon()
    startShadowsocksDaemon()
    setupIptables()

    true
  }

  def notifyForegroundAlert(title: String, info: String, visible: Boolean) {
    val openIntent = new Intent(this, classOf[Shadowsocks])
    openIntent.setFlags(Intent.FLAG_ACTIVITY_REORDER_TO_FRONT)
    val contentIntent = PendingIntent.getActivity(this, 0, openIntent, 0)
    val closeIntent = new Intent(Action.CLOSE)
    val actionIntent = PendingIntent.getBroadcast(this, 0, closeIntent, 0)
    val builder = new NotificationCompat.Builder(this)

    builder
      .setWhen(0)
      .setColor(ContextCompat.getColor(this, R.color.material_accent_500))
      .setTicker(title)
      .setContentTitle(getString(R.string.app_name))
      .setContentText(info)
      .setContentIntent(contentIntent)
      .setSmallIcon(R.drawable.ic_stat_shadowsocks)
      .addAction(android.R.drawable.ic_menu_close_clear_cancel, getString(R.string.stop),
        actionIntent)

    if (visible)
      builder.setPriority(NotificationCompat.PRIORITY_DEFAULT)
    else
      builder.setPriority(NotificationCompat.PRIORITY_MIN)

    startForeground(1, builder.build)
  }

  def onBind(intent: Intent): IBinder = {
    Log.d(TAG, "onBind")
    if (Action.SERVICE == intent.getAction) {
      binder
    } else {
      null
    }
  }

  override def onCreate() {
    super.onCreate()
    ConfigUtils.refresh(this)
  }

  def killProcesses() {
    try {
      Process.killProcess(sslocalPid)
    } catch {
      case e: Throwable => Log.e(TAG, "unable to kill ss-local")
    }
    try {
      Process.killProcess(sstunnelPid)
    } catch {
      case e: Throwable => Log.e(TAG, "unable to kill ss-tunnel")
    }
    try {
      Process.killProcess(redsocksPid)
    } catch {
      case e: Throwable => Log.e(TAG, "unable to kill redsocks")
    }
    try {
      Process.killProcess(pdnsdPid)
    } catch {
      case e: Throwable => Log.e(TAG, "unable to kill pdnsd")
    }

    Console.runRootCommand(Utils.getIptables + " -t nat -F OUTPUT")
  }

  def setupIptables() = {
    val init_sb = new ArrayBuffer[String]
    val http_sb = new ArrayBuffer[String]

    init_sb.append("ulimit -n 4096")
    init_sb.append(Utils.getIptables + " -t nat -F OUTPUT")

    val cmd_bypass = Utils.getIptables + CMD_IPTABLES_RETURN
    if (!InetAddress.getByName(config.proxy.toUpperCase).isInstanceOf[Inet6Address]) {
      init_sb.append(cmd_bypass.replace("-p tcp -d 0.0.0.0", "-d " + config.proxy))
    }
    init_sb.append(cmd_bypass.replace("-p tcp -d 0.0.0.0", "-d 127.0.0.1"))
    init_sb.append(cmd_bypass.replace("-p tcp -d 0.0.0.0", "-m owner --uid-owner " + myUid))
    init_sb.append(cmd_bypass.replace("-d 0.0.0.0", "--dport 53"))

    init_sb.append(Utils.getIptables
      + " -t nat -A OUTPUT -p udp --dport 53 -j DNAT --to-destination 127.0.0.1:" + 8153)

    if (!config.isProxyApps || config.isBypassApps) {
      http_sb.append(Utils.getIptables + CMD_IPTABLES_DNAT_ADD_SOCKS)
    }
    if (config.isProxyApps) {
      if (apps == null || apps.length <= 0) {
        apps = AppManager.getProxiedApps(this, config.proxiedAppString)
      }
      val uidSet: mutable.HashSet[Int] = new mutable.HashSet[Int]
      for (app <- apps) {
        if (app.proxied) {
          uidSet.add(app.uid)
        }
      }
      for (uid <- uidSet) {
        if (!config.isBypassApps) {
          http_sb.append((Utils.getIptables + CMD_IPTABLES_DNAT_ADD_SOCKS).replace("-t nat", "-t nat -m owner --uid-owner " + uid))
        } else {
          init_sb.append(cmd_bypass.replace("-d 0.0.0.0", "-m owner --uid-owner " + uid))
        }
      }
    }
    Console.runRootCommand(init_sb.toArray)
    Console.runRootCommand(http_sb.toArray)
  }

  override def startRunner(c: Config) {

    TrafficMonitor.reset()

    config = c

    // register close receiver
    val filter = new IntentFilter()
    filter.addAction(Intent.ACTION_SHUTDOWN)
    filter.addAction(Action.CLOSE)
    closeReceiver = (context: Context, intent: Intent) => {
      Toast.makeText(context, R.string.stopping, Toast.LENGTH_SHORT).show()
      stopRunner()
    }
    registerReceiver(closeReceiver, filter)

    if (Utils.isLollipopOrAbove) {
      val screenFilter = new IntentFilter()
      screenFilter.addAction(Intent.ACTION_SCREEN_ON)
      screenFilter.addAction(Intent.ACTION_SCREEN_OFF)
      screenFilter.addAction(Intent.ACTION_USER_PRESENT)
      lockReceiver = (context: Context, intent: Intent) => if (getState == State.CONNECTED) {
        val action = intent.getAction
        if (action == Intent.ACTION_SCREEN_OFF) {
          notifyForegroundAlert(getString(R.string.forward_success),
            getString(R.string.service_running).formatLocal(Locale.ENGLISH, config.profileName), false)
        } else if (action == Intent.ACTION_SCREEN_ON) {
          val keyGuard = getSystemService(Context.KEYGUARD_SERVICE).asInstanceOf[KeyguardManager]
          if (!keyGuard.inKeyguardRestrictedInputMode) {
            notifyForegroundAlert(getString(R.string.forward_success),
              getString(R.string.service_running).formatLocal(Locale.ENGLISH, config.profileName), true)
          }
        } else if (action == Intent.ACTION_USER_PRESENT) {
          notifyForegroundAlert(getString(R.string.forward_success),
            getString(R.string.service_running).formatLocal(Locale.ENGLISH, config.profileName), true)
        }
      }
        registerReceiver(lockReceiver, screenFilter)
    }

    ShadowsocksApplication.track(TAG, "start")

    changeState(State.CONNECTING)

    Future {

      if (config.proxy == "198.199.101.152") {
        val holder = ShadowsocksApplication.containerHolder
        try {
          config = ConfigUtils.getPublicConfig(getBaseContext, holder.getContainer, config)
        } catch {
          case ex: Exception =>
            changeState(State.STOPPED, getString(R.string.service_failed))
            stopRunner()
            config = null
        }
      }

      if (config != null) {

        // Clean up
        killProcesses()

        var resolved: Boolean = false
        if (!Utils.isNumeric(config.proxy)) {
          Utils.resolve(config.proxy, enableIPv6 = true) match {
            case Some(a) =>
              config.proxy = a
              resolved = true
            case None => resolved = false
          }
        } else {
          resolved = true
        }

        if (resolved && handleConnection) {

          // Set DNS
          flushDns()

          notifyForegroundAlert(getString(R.string.forward_success),
            getString(R.string.service_running).formatLocal(Locale.ENGLISH, config.profileName), true)
          changeState(State.CONNECTED)
        } else {
          changeState(State.STOPPED, getString(R.string.service_failed))
          stopRunner()
        }
      }
    }
  }

  override def stopRunner() {

    TrafficMonitor.reset()

    // channge the state
    changeState(State.STOPPING)

    // clean up recevier
    if (closeReceiver != null) {
      unregisterReceiver(closeReceiver)
      closeReceiver = null
    }

    if (Utils.isLollipopOrAbove) {
      if (lockReceiver != null) {
        unregisterReceiver(lockReceiver)
        lockReceiver = null
      }
    }

    ShadowsocksApplication.track(TAG, "stop")

    // reset NAT
    killProcesses()

    // stop the service if no callback registered
    if (getCallbackCount == 0) {
      stopSelf()
    }

    stopForeground(true)

    // change the state
    changeState(State.STOPPED)
  }

  override def stopBackgroundService() {
    stopSelf()
  }

  override def getTag = TAG
  override def getServiceMode = Mode.NAT
  override def getContext = getBaseContext
}
